""" functions to load genes, and identify transcripts containing de novos.
"""

from denovonear.transcript import Transcript


def get_deprecated_gene_ids(filename):
    """ gets a dict of the gene IDs used during in DDD datasets that have been
    deprecated in favour of other gene IDs
    """
    
    deprecated = {}
    with open(filename) as handle:
        for line in handle:
            line = line.strip().split()
            old = line[0]
            new = line[1]
            deprecated[old] = new
    
    return deprecated

def get_transcript_lengths(ensembl, transcript_ids):
    """ finds the protein length for ensembl transcript IDs for a gene
    
    Args:
        ensembl: EnsemblRequest object to request sequences and data
            from the ensembl REST API
        transcript_ids: list of transcript IDs for a single gene
    
    Returns:
        dictionary of lengths (in amino acids), indexed by transcript IDs
    """
    
    transcripts = {}
    for transcript_id in transcript_ids:
        # get the transcript's protein sequence via the ensembl REST API
        try:
            seq = ensembl.get_protein_seq_for_transcript(transcript_id)
        except ValueError:
            continue
        
        transcripts[transcript_id] = len(seq)
    
    return transcripts

def construct_gene_object(ensembl, transcript_id):
    """ creates an Transcript object for a gene from ensembl databases
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        transcript_id: string for an Ensembl transcript ID
    
    Returns:
        a Transcript object, containing transcript coordinates and gene and
        transcript sequence.
    
    Raises:
        ValueError if CDS from genomic sequence given gene coordinates and CDS
        retrieved from Ensembl do not match.
    """
    
    # get the sequence for the identified transcript
    (chrom, start, end, strand, genomic_sequence) = ensembl.get_genomic_seq_for_transcript(transcript_id, expand=10)
    cds_sequence = ensembl.get_cds_seq_for_transcript(transcript_id)
    
    # get the locations of the exons and cds from ensembl
    cds_ranges = ensembl.get_cds_ranges_for_transcript(transcript_id)
    exon_ranges = ensembl.get_exon_ranges_for_transcript(transcript_id)
    
    # start a Transcript object with the locations and sequence
    transcript = Transcript(transcript_id, chrom, start, end, strand)
    transcript.set_exons(exon_ranges, cds_ranges)
    transcript.set_cds(cds_ranges)
    
    transcript.add_cds_sequence(cds_sequence)
    transcript.add_genomic_sequence(genomic_sequence, offset=10)
    
    return transcript

def get_de_novos_in_transcript(transcript, de_novos):
    """ get the de novos within the coding sequence of a transcript
    
    Args:
        transcript: Transcript object, which defines the transcript coordinates
        de_novos: list of chromosome sequence positions for de novo events
    
    Returns:
        list of de novo positions found within the transcript
    """
    
    in_transcript = []
    for de_novo in de_novos:
        # we check if the de novo is within the transcript by converting the
        # chromosomal position to a CDS-based position. Variants outside the CDS
        # will raise an error, which we catch and pass on. It's better to do
        # this, rather than use the function in_coding_region(), since that
        # function does not allow for splice site variants.
        try:
            transcript.chrom_pos_to_cds(de_novo)
            in_transcript.append(de_novo)
        except RuntimeError:
            continue
    
    return in_transcript
    
def get_transcript_ids(ensembl, gene_id):
    """ gets transcript IDs for a gene.
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        gene_id: HGNC symbol for gene
    
    Returns:
        dictionary of transcript ID: transcript lengths for all transcripts
        for a given HGNC symbol.
    """
    
    ensembl_genes = ensembl.get_genes_for_hgnc_id(gene_id)
    transcript_ids = ensembl.get_transcript_ids_for_ensembl_gene_ids(ensembl_genes, [gene_id])
    
    # sometimes we get HGNC symbols that do not match the ensembl rest version
    # that we are currentl using. We can look for earlier HGNC symbols for
    # the gene using the service at rest.genenames.org
    alt_symbols = []
    if len(transcript_ids) == 0:
        alt_symbols = ensembl.get_previous_symbol(gene_id)
        genes = [ensembl.get_genes_for_hgnc_id(symbol) for symbol in alt_symbols]
        genes = [item for sublist in genes for item in sublist]
        ensembl_genes += genes
        symbols = [gene_id] + alt_symbols
        
        transcript_ids = ensembl.get_transcript_ids_for_ensembl_gene_ids(ensembl_genes, symbols)
    
    return get_transcript_lengths(ensembl, transcript_ids)

def load_gene(ensembl, gene_id, de_novos=[]):
    """ sort out all the necessary sequences and positions for a gene
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        gene_id: HGNC symbol for gene
        de_novos: list of de novo positions, so we can check they all fit in
            the gene transcript
        
    Returns:
        list of Transcript objects for gene, including genomic ranges and sequences
    """
    
    transcripts = minimise_transcripts(ensembl, gene_id, de_novos)
    
    genes = []
    for transcript_id in transcripts:
        gene = construct_gene_object(ensembl, transcript_id)
        genes.append(gene)
    
    if len(genes) == 0:
        raise IndexError("{0}: no suitable transcripts".format(gene_id))
    
    return genes
    
def count_de_novos_per_transcript(ensembl, gene_id, de_novos=[]):
    """ count de novos in transcripts for a gene.
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        gene_id: HGNC symbol for gene
        de_novos: list of de novo positions, so we can check they all fit in
            the gene transcript
        
    Returns:
        dictionary of lengths and de novo counts, indexed by transcript IDs.
    """
    
    transcripts = get_transcript_ids(ensembl, gene_id)
    
    # TODO: allow for genes without any coding sequence.
    if len(transcripts) == 0:
        raise IndexError("{0} lacks coding transcripts".format(gene_id))
    
    # count the de novos observed in all transcripts
    counts = {}
    for key in transcripts:
        try:
            gene = construct_gene_object(ensembl, key)
            total = len(get_de_novos_in_transcript(gene, de_novos))
            if total > 0:
                counts[key] = {}
                counts[key]["n"] = total
                counts[key]["len"] = transcripts[key]
        except ValueError:
            pass
    
    return counts

def minimise_transcripts(ensembl, gene_id, de_novos):
    """ get a set of minimal transcripts to contain all the de novos.
    
    We identify the minimal number of transcripts to contain all de novos. This
    allows for de novos on mutually exclusive transcripts. The transcripts are
    selected on the basis of containing the most number of de novos, while also
    being the longest possible transcript for the gene.
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        gene_id: HGNC symbol for gene
        de_novos: set of de novo positions
    
    Returns:
        dictionary of lengths and de novo counts, indexed by transcript ID for
        the set of minimal transcripts necessary to contain all de novos.
    """
    
    if len(de_novos) == 0:
        return {}
    
    counts = count_de_novos_per_transcript(ensembl, gene_id, de_novos)
    
    if len(counts) == 0:
        return {}
    
    # find the transcripts with the most de novos
    max_count = max( val["n"] for key, val in counts.items() )
    transcripts = [ key for key, val in counts.items() if val["n"] == max_count ]
    
    # find the transcripts with the greatest length, and the most de novos
    max_length = max( counts[key]["len"] for key in transcripts )
    tx_ids = [ key for key in transcripts if counts[key]["len"] == max_length ]
    max_transcripts = {x: counts[x] for x in counts if x in tx_ids}
    
    # find which de novos occur in the transcript with the most de novos
    gene = construct_gene_object(ensembl, next(iter(max_transcripts)))
    denovos_in_gene = get_de_novos_in_transcript(gene, de_novos)
    
    # trim the de novos to the ones not in the current transcript
    leftovers = [ x for x in de_novos if x not in denovos_in_gene ]
    
    # and recursively return the transcripts in the current transcript, along
    # with transcripts for the reminaing de novos
    max_transcripts.update(minimise_transcripts(ensembl, gene_id, leftovers))
    
    return max_transcripts

    



"""
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from denovonear.site_specific_rates import get_mutated_aa

class Consequences(object):
    """ class to identify HGVS-like codes for variants
    
    This class relies on a Transcript object: self.transcript, which defines
    exon positions and genomic sequence to get coding positions, codon sequences
    and translations.
    """
    
    def get_conseqence(self, pos, ref, alt):
        """ get a consequence string for a de novo (e.g. R226Q)
        
        Args:
            pos: nucleotide position of the variant (chromosome coordinates)
            ref: reference allele for the variant
            alt: alternate allele of the variant
        
        Returns:
            string denoting the amino acid change at the given residue (e.g
            R226Q for an arginine subsituted to a glutamine at the 226th
            amino acid).
        """
        
        # make sure the alleles match the strand of the gene
        if self.transcript.get_strand() == "-":
            ref = self.transcript.reverse_complement(ref)
            alt = self.transcript.reverse_complement(alt)
        
        pos = self._adjust_position(pos, ref, alt)
        
        if not self.transcript.in_coding_region(pos):
            return self._get_splice_consequence(pos, ref, alt)
        
        if len(alt) != len(ref):
            return self._get_indel_consequence(pos, ref, alt)
        
        return self._get_missense_consequence(pos, ref, alt)
    
    def _adjust_position(self, pos, ref, alt):
        """ adjust the chromosome position for the matched sequence of alleles.
        
        Sometimes we have a ref allele of TGGGG, and an alt allele of TGGGGGG.
        We adjust the nucleotide position for the section of the alleles that
        match, since otherwise we won't have the position for the amino acid
        that is being altered.
        
        Args:
            pos: nucleotide position of the variant (chromosome coordinates)
            ref: reference allele for the variant
            alt: alternate allele of the variant
        
        Returns:
            position adjusted by the number of bases of the alleles that match.
        """
        
        # if we are on the reverse strand, then reverse the alleles, so as to
        # get them back oriented to the genomic strand. Reverse strand alleles
        # remain as the  complement
        if self.transcript.get_strand() == "-":
            ref = ref[::-1]
            alt = alt[::-1]
        
        matched = 0
        try:
            while ref[matched] == alt[matched]:
                matched += 1
        except IndexError:
            pass
        
        if self.transcript.get_strand() == "-" and len(ref) > len(alt):
            pos += len(ref) - 2
        else:
            pos += matched
        
        return pos
    
    def _get_indel_consequence(self, pos, ref, alt):
        """ figure out the HGVS-like code for indels (frameshifts too) variants.
        
        TODO: This function currently does not provide the full indel
        information. For example, frameshift variants should also have the
        distance in residues to the next stop codon (assuming the frameshift
        variant has introduced a nearby stop codon). Deletions and insertions
        should not just list the initial amino acid that differs, they should
        also include the amino acids that have been removed (or introduced).
        
        Args:
            pos: nucleotide position of the variant (chromosome coordinates)
            ref: reference allele for the variant
            alt: alternate allele of the variant
        
        Returns:
            HGVS-like code for a indel variant e.g. R226fs.
        """
        
        codon = self.transcript.get_codon_info(pos)
        
        # change the codon number from being 0-based to 1-based
        codon["codon_number"] += 1
        
        # the frameshift variants are ones where the difference in length
        # between the ref and alt alleles is not divisible by 3
        if abs(len(ref) - len(alt)) % 3 != 0:
            return "{}{}fs".format(codon["initial_aa"], codon["codon_number"])
        elif len(ref) > len(alt):
            return "{}{}del".format(codon["initial_aa"], codon["codon_number"])
        elif len(ref) < len(alt):
            return "{}{}ins".format(codon["initial_aa"], codon["codon_number"])
    
    def _get_splice_consequence(self, pos, ref, alt):
        """ figure out the HGVS-like code for splice site variants.
        
        First find the closest exon, then the code is based off the distance to
        the nearest end. We also need to take the strand into account. See
        http://www.hgvs.org/mutnomen/recs-DNA.html
        
        Args:
            pos: nucleotide position of the variant (chromosome coordinates)
            ref: reference allele for the variant
            alt: alternate allele of the variant
        
        Returns:
            HGVS-like code for a splice site variant e.g. c.1000-2G>A,
            indicating an exon boundary at position 1000 of the transcript,
            where the variant is 2 bp upstream, and the variant is a G
            substituted with an A.
        """
        
        start, end = self.transcript.find_closest_exon(pos)
        
        distance = min(abs(end - pos), abs(start - pos))
        if pos < start:
            cds_pos = self.transcript.chrom_pos_to_cds(start)['pos']
            separator = "-"
        else:
            separator = "+"
            cds_pos = self.transcript.chrom_pos_to_cds(end)['pos']
        
        cds_pos += 1
        if self.transcript.get_strand() == "-":
            distance -= 1
            separator = {"-": "+", "+": "-"}[separator]
        
        return "c.{}{}{}{}>{}".format(cds_pos, separator, distance, ref, alt)
    
    def _get_missense_consequence(self, pos, ref, alt):
        """ get HGVS-like code for missense variants
        
        Args:
            pos: nucleotide position of the variant (chromosome coordinates)
            ref: reference allele for the variant
            alt: alternate allele of the variant
        
        Returns:
            HGVS-like code for a missense variant e.g. R226Q for an arginine at
            the 226nd amino acid being substituted with a glutamine.
        """
        
        codon = self.transcript.get_codon_info(pos)
        
        # change the codon number from being 0-based to 1-based
        codon["codon_number"] += 1
        
        mutated_aa = get_mutated_aa(self.transcript, alt, codon["codon_seq"],
            codon["intra_codon"])
        
        return "{}{}{}".format(codon["initial_aa"], codon["codon_number"], mutated_aa)



""" class to plot protein domains.
"""

from __future__ import division

import random

import webcolors

class DomainPlot(object):
    """ class to plot protein domains.
    """
    
    def plot_domains(self, domains, sequence):
        """ plots protein domains
        
        Args:
            domains: list of dictionaries for different domain annotations in a
                protein
            sequence: protein sequence as string
        """
        
        length = len(sequence) / self.size
        
        # increment the y_offset position, so as to avoid overplotting between
        # gene, transcript and protein diagrams
        self.y_offset += self.box_height * 3
        
        # make sure the full protein is visible
        self.add_box(x_pos=0, width=self.size, fillcolor="white")
        
        for domain in domains:
            self.plot_single_domain(domain, length)
        
        # get the initial coordinates to place the de novos at
        for x in self.de_novos:
            self.de_novos[x]["coordinate"] = ()
            cds = self.transcript.chrom_pos_to_cds(self.de_novos[x]["start_pos"])['pos']
            codon = self.transcript.get_codon_number_for_cds_position(cds)
            offset = self.transcript.get_position_within_codon(cds)/3
            
            position = (codon + offset)/length
            width = max(0.33/length, self.size/1000)
            self.de_novos[x]["coordinate"] = (position, width)
        
        self.add_de_novos(self.de_novos)
        
        # and include the gene symbol and amino acid length on the domain plot
        text = "{} ({} aa)".format(self.hgnc_symbol, len(sequence))
        self.add_text(self.size, text, y_adjust=self.box_height*1.5, horizontalalignment="right")
    
    def plot_single_domain(self, domain, length):
        """ plots a single domain
        """
        
        x_pos = domain["start"] / length
        width = (domain["end"] - domain["start"]) / length
        
        color = random.choice(self.colorset)
        
        # add a box on the domain plot, as well as a text label centered
        # below the box
        self.add_box(x_pos, width, fillcolor=color)
        self.add_text(x_pos + self.box_height / 4, domain["domain_type"], \
            y_adjust=self.box_height, horizontalalignment="right", \
            rotate=315, color=color)
       



""" class to plot gene diagrams. This is a base class for plotting genes,
transcripts, and proteins (with domains).
"""

from __future__ import division
from __future__ import print_function

import math

import webcolors
import cairocffi as cairo

from denovonear.gene_plot.domains import DomainPlot
from denovonear.gene_plot.transcript import TranscriptPlot
from denovonear.gene_plot.genomic import GenomicPlot
from denovonear.gene_plot.consequences import Consequences

class DiagramPlotter(GenomicPlot, TranscriptPlot, DomainPlot, Consequences):
    
    # Define a set of colors that we can plot domains as, but don't allow any
    # colors close to white, as they are difficult to distingush from background
    global white
    white = ["aliceblue", "azure", "beige", "cornsilk", "floralwhite",
        "gainsboro", "ghostwhite", "honeydew", "ivory", "lavenderblush", "linen",
        "lightyellow", "mintcream", "oldlace", "seashell", "snow", "white",
        "whitesmoke"]
    colorset = webcolors.css3_names_to_hex.keys()
    colorset = [ x for x in colorset if x not in white ]
    
    size = 1000
    y_offset = 0
    box_height = size/20
    
    def __init__(self, transcript, hgnc_symbol, de_novos, filename=None):
        """ start the class with a matplotlib plot
        
        Args:
            transcript: Transcript object for the gene
            hgnc_symbol: hgnc_symbol for the gene
            de_novos: list of genomic coordinates for de novo variants
        """
        
        self.transcript = transcript
        self.de_novos = de_novos
        self.hgnc_symbol = hgnc_symbol
        
        if transcript.get_strand() == "-":
            for key in self.de_novos:
                self.de_novos[key]["start_pos"] -= 1
        
        if filename is None:
            filename = "gene_plot.{0}.pdf".format(self.hgnc_symbol)
        
        self.surface = cairo.PDFSurface(filename,  self.size,  self.size/1.5)
        self.cr = cairo.Context(self.surface)
    
    def add_box(self, x_pos, width, y_adjust=0, height=None, strokecolor=None, fillcolor=None, linewidth=None):
        """ adds a rectangle to the plot
        """
        
        if height is None:
            height = self.box_height
        
        if strokecolor is None:
            # default to a black stroke
            strokecolor = [0, 0, 0]
        else:
            strokecolor = webcolors.name_to_rgb(strokecolor, spec=u'css3')
            strokecolor = [ x/255 for x in strokecolor ]
        
        if fillcolor is None:
            # default to a white fill
            fillcolor = [1, 1, 1]
        else:
            fillcolor = webcolors.name_to_rgb(fillcolor, spec=u'css3')
            fillcolor = [ x/255 for x in fillcolor ]
        
        if linewidth is None:
            linewidth = self.size/200
        
        self.cr.set_line_width(linewidth)
        self.cr.set_source_rgb(*strokecolor)
        self.cr.rectangle(x_pos, self.y_offset + y_adjust, width, height)
        self.cr.stroke_preserve()
        self.cr.set_source_rgb(*fillcolor)
        self.cr.fill()
    
    def add_text(self, x_pos, text, y_adjust=0, horizontalalignment=None, rotate=None, fontsize=None, color=None):
        """ adds some text to the plot
        
        Args:
            x_pos: horizontal position at which to plot the text
            text: string of text to be plotted
            y_adjust: how far away to plot the text from the current y_offset
        """
        
        y_pos = self.y_offset + y_adjust
        
        if fontsize is None:
            fontsize = "medium"
        
        sizes = {"small": self.size/80, "medium": self.size/50, "large": self.size/30}
        
        self.cr.select_font_face("Arial")
        self.cr.set_font_size(sizes[fontsize])
        px = max(self.cr.device_to_user_distance(1, 1))
        fascent, fdescent, fheight, fxadvance, fyadvance = self.cr.font_extents()
        xbearing, ybearing, width, height, xadvance, yadvance = \
                self.cr.text_extents(text)
        
        if rotate is not None:
            x_rot_height = width * math.sin(math.radians(rotate))
            x_rot_width = math.sqrt(width ** 2 - x_rot_height ** 2)
            rot_x_delta = width - x_rot_width
            rot_y_delta = x_rot_height - height
        
        if horizontalalignment is not None:
            if horizontalalignment == "center":
                x_pos = x_pos - xbearing - width / 2
            elif horizontalalignment == "right":
                x_pos = x_pos - xbearing - width
            if rotate:
                x_pos = x_pos + rot_x_delta
                y_pos = y_pos - rot_y_delta
    
        y_pos = y_pos - fdescent + fheight / 2
        
        self.cr.move_to(x_pos, y_pos)
        
        if rotate is not None:
            self.cr.rotate(math.radians(rotate))
        
        if color is None:
            # default to a black fill
            color = [0, 0, 0]
        else:
            color = webcolors.name_to_rgb(color, spec=u'css3')
            color = [ x/255 for x in color ]
        
        self.cr.set_source_rgb(*color)
        self.cr.show_text(text)
        
        if rotate is not None:
            self.cr.rotate(-math.radians(rotate))
    
    def add_de_novos(self, de_novos):
        """ adds a line to show the position of a de novo
        
        Args:
            de_novos: dictionary of variants, indexed by (position, person_id)
                tuples. Each variant entry contains a "coordinate" entry, which
                is a tuple of (x_position, width).
        """
        
        height = self.box_height / 2
        
        i = -1
        for key in sorted(de_novos):
            i += 1
            color = "gray"
            if de_novos[key]["source"] == "internal":
                color = "red"
            
            self.add_box(*de_novos[key]["coordinate"], height=height, y_adjust=-height, \
                fillcolor=color, strokecolor=color, linewidth=0)
            
            (x, y) = self.rotate_indicator(de_novos, key, i, color, height)
                
            text = self.get_conseqence(de_novos[key]["start_pos"], \
                de_novos[key]["ref_allele"], de_novos[key]["alt_allele"])
            
            self.add_text(x, text, y, rotate=315, fontsize="small")
    
    def rotate_indicator(self, de_novos, key, i, color, height):
        """
        """
        
        (x_pos, width) = de_novos[key]["coordinate"]
        positions = self.get_overlap_positions(de_novos, key)
        if len(positions) == 1:
            return (x_pos, -height)
        
        radians = self.get_rotation(positions, i)
        dy = height * math.sin(radians)
        dx = math.sqrt(height**2 - dy**2)
        
        if math.degrees(radians) < 90:
            dx = -dx
        
        # get the rgb values for the required stroke color
        strokecolor = webcolors.name_to_rgb(color, spec=u'css3')
        strokecolor = [ x/255 for x in strokecolor ]

        # plot a line arcing out from the overlap point
        self.cr.new_path()
        self.cr.move_to(x_pos, self.y_offset - height)
        self.cr.rel_line_to(dx, -dy)
        self.cr.set_line_width(self.size/1000)
        self.cr.set_source_rgb(*strokecolor)
        self.cr.stroke()
        
        return (x_pos + dx, -height-dy)
    
    def get_overlap_positions(self, de_novos, key):
        """ identify the variants which overlap the current site
        
        Args:
            de_novos: list of de novos within gene
            key: current de novo to check
        
        Returns:
            list positions for overlapping variants
        """
        
        (x_pos, width) = de_novos[key]["coordinate"]
        
        # check how many other sites the de novo overlaps
        overlaps = [ x_pos <= de_novos[z]["coordinate"][0] + de_novos[z]["coordinate"][1] \
            and x_pos + width >= de_novos[z]["coordinate"][0] for z in sorted(de_novos) ]
        return [ x for x, value in enumerate(overlaps) if value ]
    
    def get_rotation(self, positions, i):
        """ get the rotation angle for line segments for overlapping de novos
        
        Args:
            positions: list of positions that overlap the current de novo
            i: position in list for current de novo
        
        Returns:
            angle to plot segment in radians
        """
        
        # figure out the angle for the indicator line
        increment = 180/(len(positions) + 1)
        angle = increment + positions.index(i) * increment
        return math.radians(angle)
    
    def export_figure(self):
        """ exports the plot as a pdf
        """
        
        self.cr.save()
        



""" functions to parse the output from requests to the InterPro REST API
"""

# from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys

class InterProParser(object):
    """ class to parse results from domain prediction runs from the EBI REST API
    """
    
    def find_domain_match(self, entry, domains):
        """ find matches between the current domain to the other domain results
        
        If the domain overlaps any other domain, we check if the description
        terms match, if they do, it is simply a matter of different tools
        predicting the same domain, in which case we avoid duplicating the
        domain.
        
        Args:
            entry: dictionary entry for domain prediction
        
        Returns:
            Return -1 when there are no matches, otherwise return the list
            position that the match is found at
        """
        
        match = -1
        for index in range(len(domains)):
            result = domains[index]
            if entry["end"] >= result["start"] and result["end"] >= entry["start"]:
                if len(entry["domain_type"] & result["domain_type"]) > 0:
                    match = index
                    break
        
        return match
    
    def standardise_names(self, domains):
        """ picks the shortest name from a list of alternates for each domain
        """
        
        # convert the list of alternative domain names into a single identifier
        for pos in range(len(domains)):
            names = list(domains[pos]["domain_type"])
            
            names = [ x.replace(" domain", "") for x in names ]
            name_lengths = [ len(x) for x in names ]
            
            # find the shortest name (note, this won't necessarily be the best
            # name, but it's a start).
            name_pos = name_lengths.index(min(name_lengths))
            
            domains[pos]["domain_type"] = names[name_pos]
        
        return domains
    
    def build_domain_entry(self, line):
        """ clean up a interpro results line.
        """
        
        entry = {}
            
        description = line[5].rstrip(".")
        entry["tool"] = [line[3]]
        entry["start"] = int(line[6])
        entry["end"] = int(line[7])
        
        # normalise zinc finger domain types, since these can have many alternative
        # descriptors
        if "zinc finger" in description.lower():
            description = "Zinc finger"
        
        if len(line) == 11:
            entry["domain_type"] = set([description])
        elif len(line) > 11:
            entry["domain_type"] = set([description, line[12]])
            entry["accession"] = line[11]
        
        # normalise the domain type from the "Coils" tool
        if entry["tool"] == ["Coils"]:
            entry["domain_type"] = set(["Coiled coil"])
        
        entry["domain_type"].discard("")
        
        return entry
    
    def parse_interpro_results(self, api_response):
        """ parses the output of an InterPro analysis run
        """
        
        domains = []
        for line in api_response.strip().split("\n"):
            line = line.split("\t")
            
            entry = self.build_domain_entry(line)
            
            # I'm not using PANTHER, Gene3D or SUPERFAMILY results, since Ensembl
            # doesn't include them (and because they have slightly different naming
            # conventions, whichmakes more difficult to match predictions between
            # different tools)
            if len(set(entry["tool"]) & set(["PANTHER", "Gene3D", "SUPERFAMILY",
                    "SignalP_GRAM_POSITIVE", "SignalP_GRAM_NEGATIVE"])) > 0:
                continue
            
            match = self.find_domain_match(entry, domains)
            
            if match > -1:
                # We extend the original to contain the overlaps and tool info.
                domain = domains[match]
                domains[match]["start"] = min(entry["start"], domain["start"])
                domains[match]["end"] = max(entry["end"], domain["end"])
                domains[match]["domain_type"] = entry["domain_type"] | domain["domain_type"]
                domains[match]["tool"] = list(set(entry["tool"]) | set(domain["tool"]))
            else:
                domains.append(entry)
        
        return domains



""" Script to identify transcripts containing de novos
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys
import os
import math
import argparse

from denovonear.load_gene import construct_gene_object, get_deprecated_gene_ids, \
    count_de_novos_per_transcript, minimise_transcripts
from denovonear.ensembl_requester import EnsemblRequest
from denovonear.site_specific_rates import SiteRates
from denovonear.load_de_novos import load_de_novos

def get_options():
    """ get the command line switches
    """
    
    parser = argparse.ArgumentParser(description="Identify transcripts \
        for a gene containing de novo events.")
    parser.add_argument("--de-novos", dest="input", required=True, help="Path to \
        file listing de novo variants in genes.")
    parser.add_argument("--out", dest="output", required=True, help="output \
        filename")
    parser.add_argument("--deprecated-genes", dest="deprecated_genes", \
        help="deprecated gene IDs filename")
    parser.add_argument("--genome-build", dest="genome_build", choices=["grch37",
        "GRCh37", "grch38", "GRCh38"], default="grch37", help="Genome build "+ \
        "that the de novo coordinates are based on (GrCh37 or GRCh38")
    parser.add_argument("--cache-folder", dest="cache_folder", \
        default=os.path.join(os.path.dirname(__file__), "cache"), help="folder \
        to cache Ensembl data into (defaults to clustering code directory)")
    
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--all-transcripts", action="store_true", default=False,\
        help="Flag if you want to identify all transcripts with more than" + \
        "one de novo on it.")
    group.add_argument("--minimise-transcripts", action="store_true", \
        default=False, help="Flag if you want to identify the minimal set of" +\
        "transcripts to contain all de novos.")
    
    args = parser.parse_args()
    
    return args.input, args.output, args.deprecated_genes, \
        args.cache_folder, args.genome_build.lower(), args.all_transcripts, \
        args.minimise_transcripts

def main():
    
    input_file, output_file, old_gene_id_file, cache_dir, genome_build, \
        all_transcripts, minimal_transcripts = get_options()
    
    # load all the data
    ensembl = EnsemblRequest(cache_dir, genome_build)
    
    old_gene_ids = {}
    if old_gene_id_file is not None:
        old_gene_ids = get_deprecated_gene_ids(old_gene_id_file)
    
    known_de_novos = load_de_novos(input_file, exclude_indels=False)
    
    output = open(output_file, "w")
    output.write("hgnc_symbol\ttranscript_id\tlength\tde_novos\n")
    
    for gene_id in sorted(known_de_novos):
        print(gene_id)
        de_novos = known_de_novos[gene_id]
        func_events = de_novos["missense"] + de_novos["nonsense"]
        
        # fix HGNC IDs that have been discontinued in favour of other gene IDs
        if gene_id in old_gene_ids:
            gene_id = old_gene_ids[gene_id]
        
        # find the counts per transcript, depending on whether we want to count
        # for all transcripts containing one or more de novos, or to find the
        # minimum set of transcripts to contain the de novos
        try:
            if all_transcripts:
                counts = count_de_novos_per_transcript(ensembl, gene_id, func_events)
            elif minimal_transcripts:
                counts = minimise_transcripts(ensembl, gene_id, func_events)
        except (ValueError, IndexError):
            print("error occured with {0}".format(gene_id))
            continue
        
        # write the transcript details to a file
        for key in counts:
            line = "{0}\t{1}\t{2}\t{3}\n".format(gene_id, key, counts[key]["len"], counts[key]["n"])
            output.write(line)
        
    output.close()

if __name__ == '__main__':
    main()



""" class to test the EnsemblRequest class
"""

import json
import unittest
import time

from denovonear.ensembl_requester import EnsemblRequest

ensembl = EnsemblRequest(cache_folder="cache", genome_build="grch37")

class TestEnsemblRequestPy(unittest.TestCase):
    """ unit test the EnsemblRequest class
    """
    
    def setUp(self):
        """ construct an EnsemblRequest object for unit tests
        """
        
        self.ensembl = ensembl
    
    def test_open_url(self):
        """ test that open_url() works correctly
        """
        
        headers = {"Content-Type": "application/json"}
        url = "http://rest.ensembl.org/overlap/id/ENSG00000172320?feature=gene"
        (response, status_code, headers) = self.ensembl.open_url(url, headers)
        
        response = json.loads(response)
        
        self.assertEqual(status_code, 200)
        self.assertEqual(response, [{
            "source": "ensembl_havana",
            "logic_name": "ensembl_havana_gene",
            "feature_type": "gene",
            "external_name": "OR5A1",
            "seq_region_name": "11",
            "strand": 1,
            "id": "ENSG00000172320",
            "gene_id": "ENSG00000172320",
            "version": 2,
            "assembly_name": "GRCh38",
            "description": "olfactory receptor family 5 subfamily A member 1 [Source:HGNC Symbol;Acc:HGNC:8319]",
            "end": 59444194,
            "biotype": "protein_coding",
            "start": 59443144}]
            )
    
    def test_get_genes_for_hgnc_id(self):
        """ test that get_genes_for_hgnc_id() works correctly
        """
        
        genes = self.ensembl.get_genes_for_hgnc_id("KMT2A")
        self.assertEqual(genes, ['ENSG00000118058', 'ENSG00000267910'])
    
    def test_get_previous_symbol(self):
        """ test that get_previous_symbol() works correctly
        """
        
        prev = self.ensembl.get_previous_symbol("KMT2A")
        self.assertEqual(prev, ["MLL"])
        
        # make a check for a gene with multiple documents, to check that we
        # don't raise an error
        prev = self.ensembl.get_previous_symbol("KRT16P1")
        self.assertEqual(prev, ["KRT14P"])
        
    def test_get_transcript_ids_for_ensembl_gene_ids(self):
        """ test that get_transcript_ids_for_ensembl_gene_ids() works correctly
        """
        
        hgnc = ["KMT2A", "MLL"]
        ensg = ['ENSG00000118058', 'ENSG00000267910']
        
        enst = self.ensembl.get_transcript_ids_for_ensembl_gene_ids(ensg, hgnc)
        
        self.assertEqual(set(enst), set(['ENST00000534358', 'ENST00000531904',
            'ENST00000389506', 'ENST00000354520', 'ENST00000532204',
            'ENST00000529852', 'ENST00000527869', 'ENST00000533790',
            'ENST00000392873']))
    
    def test_get_genomic_seq_for_transcript(self):
        """ check that get_genomic_seq_for_transcript() works correctly
        """
        
        seq = self.ensembl.get_genomic_seq_for_transcript("ENST00000302030", expand=0)
        
        self.assertEqual(seq, ('11', 59210617, 59211667, '+', 'CTTGTCCTTGTGGTCC'
            'ACGGGAAGCATGTCCATAACCAAAGCCTGGAACAGCTCATCAGTGACCATGTTCATCCTCCTGGGA'
            'TTCACAGACCATCCAGAACTCCAGGCCCTCCTCTTTGTGACCTTCCTGGGCATCTATCTTACCACC'
            'CTGGCCTGGAACCTGGCCCTCATTTTTCTGATCAGAGGTGACACCCATCTGCACACACCCATGTAC'
            'TTCTTCCTAAGCAACTTATCTTTCATTGACATCTGCTACTCTTCTGCTGTGGCTCCCAATATGCTC'
            'ACTGACTTCTTCTGGGAGCAGAAGACCATATCATTTGTGGGCTGTGCTGCTCAGTTTTTTTTCTTT'
            'GTCGGCATGGGTCTGTCTGAGTGCCTCCTCCTGACTGCTATGGCATACGACCGATATGCAGCCATC'
            'TCCAGCCCCCTTCTCTACCCCACTATCATGACCCAGGGCCTCTGTACACGCATGGTGGTTGGGGCA'
            'TATGTTGGTGGCTTCCTGAGCTCCCTGATCCAGGCCAGCTCCATATTTAGGCTTCACTTTTGCGGA'
            'CCCAACATCATCAACCACTTCTTCTGCGACCTCCCACCAGTCCTGGCTCTGTCTTGCTCTGACACC'
            'TTCCTCAGTCAAGTGGTGAATTTCCTCGTGGTGGTCACTGTCGGAGGAACATCGTTCCTCCAACTC'
            'CTTATCTCCTATGGTTACATAGTGTCTGCGGTCCTGAAGATCCCTTCAGCAGAGGGCCGATGGAAA'
            'GCCTGCAACACGTGTGCCTCGCATCTGATGGTGGTGACTCTGCTGTTTGGGACAGCCCTTTTCGTG'
            'TACTTGCGACCCAGCTCCAGCTACTTGCTAGGCAGGGACAAGGTGGTGTCTGTTTTCTATTCATTG'
            'GTGATCCCCATGCTGAACCCTCTCATTTACAGTTTGAGGAACAAAGAGATCAAGGATGCCCTGTGG'
            'AAGGTGTTGGAAAGGAAGAAAGTGTTTTCTTAGGTCATGCGTAGAAACTTATTTATCCAAACTGCT'
            'GGAGAATTAAACAATCCAAGCCTTCACCTCCACCTCTGCCTCAGG'))
    
    def test_get_cds_seq_for_transcript(self):
        """ check that get_cds_seq_for_transcript() works correctly
        """
        
        seq = self.ensembl.get_cds_seq_for_transcript("ENST00000302030")
        self.assertEqual(seq, 'ATGTCCATAACCAAAGCCTGGAACAGCTCATCAGTGACCATGTTCATC'
            'CTCCTGGGATTCACAGACCATCCAGAACTCCAGGCCCTCCTCTTTGTGACCTTCCTGGGCATCTAT'
            'CTTACCACCCTGGCCTGGAACCTGGCCCTCATTTTTCTGATCAGAGGTGACACCCATCTGCACACA'
            'CCCATGTACTTCTTCCTAAGCAACTTATCTTTCATTGACATCTGCTACTCTTCTGCTGTGGCTCCC'
            'AATATGCTCACTGACTTCTTCTGGGAGCAGAAGACCATATCATTTGTGGGCTGTGCTGCTCAGTTT'
            'TTTTTCTTTGTCGGCATGGGTCTGTCTGAGTGCCTCCTCCTGACTGCTATGGCATACGACCGATAT'
            'GCAGCCATCTCCAGCCCCCTTCTCTACCCCACTATCATGACCCAGGGCCTCTGTACACGCATGGTG'
            'GTTGGGGCATATGTTGGTGGCTTCCTGAGCTCCCTGATCCAGGCCAGCTCCATATTTAGGCTTCAC'
            'TTTTGCGGACCCAACATCATCAACCACTTCTTCTGCGACCTCCCACCAGTCCTGGCTCTGTCTTGC'
            'TCTGACACCTTCCTCAGTCAAGTGGTGAATTTCCTCGTGGTGGTCACTGTCGGAGGAACATCGTTC'
            'CTCCAACTCCTTATCTCCTATGGTTACATAGTGTCTGCGGTCCTGAAGATCCCTTCAGCAGAGGGC'
            'CGATGGAAAGCCTGCAACACGTGTGCCTCGCATCTGATGGTGGTGACTCTGCTGTTTGGGACAGCC'
            'CTTTTCGTGTACTTGCGACCCAGCTCCAGCTACTTGCTAGGCAGGGACAAGGTGGTGTCTGTTTTC'
            'TATTCATTGGTGATCCCCATGCTGAACCCTCTCATTTACAGTTTGAGGAACAAAGAGATCAAGGAT'
            'GCCCTGTGGAAGGTGTTGGAAAGGAAGAAAGTGTTTTCTTAG')
    
    def test_get_protein_seq_for_transcript(self):
        """ test that get_protein_seq_for_transcript() works correctly
        """
        
        seq = self.ensembl.get_protein_seq_for_transcript("ENST00000302030")
        self.assertEqual(seq, 'MSITKAWNSSSVTMFILLGFTDHPELQALLFVTFLGIYLTTLAWNLAL'
            'IFLIRGDTHLHTPMYFFLSNLSFIDICYSSAVAPNMLTDFFWEQKTISFVGCAAQFFFFVGMGLSE'
            'CLLLTAMAYDRYAAISSPLLYPTIMTQGLCTRMVVGAYVGGFLSSLIQASSIFRLHFCGPNIINHF'
            'FCDLPPVLALSCSDTFLSQVVNFLVVVTVGGTSFLQLLISYGYIVSAVLKIPSAEGRWKACNTCAS'
            'HLMVVTLLFGTALFVYLRPSSSYLLGRDKVVSVFYSLVIPMLNPLIYSLRNKEIKDALWKVLERKK'
            'VFS')
    
    def test_get_genomic_seq_for_region(self):
        """ test that get_genomic_seq_for_region() works correctly
        """
        
        # not that this test uses GRCh37 coordinates
        seq = self.ensembl.get_genomic_seq_for_region('11', 59210617, 59210637)
        self.assertEqual(seq, 'CTTGTCCTTGTGGTCCACGGG')
    
    def test_get_chrom_for_transcript(self):
        """ test that get_chrom_for_transcript() works correctly
        """
        
        chrom = self.ensembl.get_chrom_for_transcript("ENST00000534358", "KMT2A")
        self.assertEqual(chrom, "11")
    
    def test_get_exon_ranges_for_transcript(self):
        """ test that get_exon_ranges_for_transcript() works correctly
        """
        
        exons = self.ensembl.get_exon_ranges_for_transcript("ENST00000534358")
        self.assertEqual(exons, [(118307205, 118307659), (118339490, 118339559),
            (118342377, 118345030), (118347520, 118347697), (118348682, 118348916),
            (118350889, 118350953), (118352430, 118352807), (118353137, 118353210),
            (118354898, 118355029), (118355577, 118355690), (118359329, 118359475),
            (118360507, 118360602), (118360844, 118360964), (118361911, 118362033),
            (118362459, 118362643), (118363772, 118363945), (118365003, 118365113),
            (118365409, 118365482), (118366415, 118366608), (118366976, 118367082),
            (118368651, 118368788), (118369085, 118369243), (118370018, 118370135),
            (118370550, 118370628), (118371702, 118371862), (118372387, 118372572),
            (118373113, 118377361), (118378244, 118378324), (118379851, 118379915),
            (118380663, 118380833), (118382666, 118382740), (118390333, 118390507),
            (118390672, 118390779), (118391517, 118391600), (118392003, 118392132),
            (118392612, 118397539)])
    
    def test_get_cds_ranges_for_transcript(self):
        """ tets that get_cds_ranges_for_transcript() works correctly
        """
        
        cds = self.ensembl.get_cds_ranges_for_transcript("ENST00000534358")
        self.assertEqual(cds, [(118307228, 118307659), (118339490, 118339559),
            (118342377, 118345030), (118347520, 118347697), (118348682, 118348916),
            (118350889, 118350953), (118352430, 118352807), (118353137, 118353210),
            (118354898, 118355029), (118355577, 118355690), (118359329, 118359475),
            (118360507, 118360602), (118360844, 118360964), (118361911, 118362033),
            (118362459, 118362643), (118363772, 118363945), (118365003, 118365113),
            (118365409, 118365482), (118366415, 118366608), (118366976, 118367082),
            (118368651, 118368788), (118369085, 118369243), (118370018, 118370135),
            (118370550, 118370628), (118371702, 118371862), (118372387, 118372572),
            (118373113, 118377361), (118378244, 118378324), (118379851, 118379915),
            (118380663, 118380833), (118382666, 118382740), (118390333, 118390507),
            (118390672, 118390779), (118391517, 118391600), (118392003, 118392132),
            (118392612, 118392887)])
    
    def test_rate_limit_ensembl_requests(self):
        """ test that rate_limit_ensembl_requests() works correctly
        """
        
        current_time = time.time()
        self.ensembl.prior_time = current_time
        
        self.ensembl.rate_limit_ensembl_requests()
        delta = self.ensembl.prior_time - current_time
        
        self.assertTrue(delta >= self.ensembl.rate_limit)



""" Script to generate mutation rates based on local sequence context rates
for Ensembl transcript IDs.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys
import os
import copy
import math
import argparse
import tempfile

from denovonear.load_gene import construct_gene_object
from denovonear.ensembl_requester import EnsemblRequest
from denovonear.load_mutation_rates import load_mutation_rates
from denovonear.site_specific_rates import SiteRates

def get_options():
    """ get the command line switches
    """
    
    parser = argparse.ArgumentParser(description="determine mutation rates \
        for genes given transcript IDs.")
    
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--transcripts", dest="transcript_input",
        help="Path to file listing Ensembl transcript IDs, one ID per line.")
    group.add_argument("--genes", dest="gene_input", help="Path to file" + \
        " listing HGNC symbols, with one or more transcript IDs per gene. " + \
        "The tab-separated input format is gene symbol followed by transcript " + \
        "ID. Alternative transcripts are listed on separate lines.")
    
    parser.add_argument("--out", dest="output", required=True, help="output \
        filename")
    parser.add_argument("--rates", dest="mut_rates",
        help="Path to file containing sequence-context based mutation rates.")
    parser.add_argument("--genome-build", dest="genome_build", choices=["grch37",
        "GRCh37", "grch38", "GRCh38"], default="grch37", help="Genome build "+ \
        "that the de novo coordinates are based on (currently GRCh37 or GRCh38).")
    parser.add_argument("--cache-folder", dest="cache_folder", \
        default=os.path.join(os.path.dirname(__file__), "cache"), help="folder \
        to cache Ensembl data into (defaults to clustering code directory).")
    
    args = parser.parse_args()
    
    return args.transcript_input, args.gene_input, args.output, args.mut_rates, \
        args.cache_folder, args.genome_build.lower()

def load_transcripts(path):
    """ load a file listing transcript IDs per line
    
    Args:
        path: path to file containing transcript IDs, one per line
    
    Returns:
        list of transcript IDs eg ["ENST00000315684", "ENST00000485511"]
    """
    
    transcripts = {}
    with open(path, "r") as f:
        for line in f:
            transcripts[line.strip()] = [line.strip()]
            
    return transcripts
    
def load_genes(path):
    """ load a file listing gene and transcript IDs
    
    Eeach gene can have one or more transcript IDs associated with it, so we
    build a dictionary, indexed by HGNC symbols, and for each gene entry, retain
     a list of the possible transcript IDs.
    
    Args:
        path: path to file containing gene and transcript IDs, with each unique
            transcript ID for a gene on different lines eg
            
            gene_1    transcript_1.1    length_1    denovo_count
            gene_1    transcript_1.2    length_2    denovo_count
            gene_1    transcript_1.2    length_3    denovo_count
            gene_2    transcript_2.1    length_3    denovo_count
    
    Returns:
        list of transcript IDs eg ["ENST00000315684", "ENST00000485511"]
    """
    
    transcripts = {}
    with open(path, "r") as f:
        for line in f:
            if line.startswith("hgnc"):
                continue
            
            line = line.strip().split("\t")
            gene_id = line[0]
            transcript_id = line[1]
            
            if gene_id not in transcripts:
                transcripts[gene_id] = []
            
            transcripts[gene_id].append(transcript_id)
            
    return transcripts

def get_mutation_rates(gene_id, transcripts, mut_dict, ensembl):
    """ determines missense, nonsense and synonymous mutation rates for a gene
    
    This can estimate a mutation rate from the union of transcripts for a gene.
    This is a biased estimate of the mutation rate, where the mutation rate
    estimates is biased towards the rate from the first-ranked transcripts,
    which I prioritise by how many de novos they contain, and how long the
    coding sequence is.
    
    This isn't a problem when different transcripts have the same coding
    sequence within their shared regions, as the rates will come outthe same,
    but may differ two transcript share an overlapping region, but not in the
    same frame, so that the sites that are missense, and nonsense will differ
    between transcripts, and thus would produce different estimates of the
    mutation rate.
    
    Args:
        gene_id: ID for the current gene (can be a transcript ID, if we are
            examining single transcripts only, or can be a HGNC ID, if we are
            examining the union of mutation rates from multiple transcripts for
            a single gene).
        transcripts: dictionary of transcripts for a gene, indexed by gene_id
        mut_dict: dictionary of local sequence context mutation rates
        ensembl: EnsemblRequest object, to retrieve information from Ensembl.
    
    Returns:
        tuple of (missense, nonsense, synonymous) mutation rates
    """
    
    missense = 0
    nonsense = 0
    splice_lof = 0
    splice_region = 0
    synonymous = 0
    combined_transcript = None
    
    for transcript_id in transcripts[gene_id]:
        
        # get the gene coordinates, sequence etc, but if the transcript is
        # unusable (hence raises an error), simply move to the next transcript
        try:
            transcript = construct_gene_object(ensembl, transcript_id)
        except ValueError:
            continue
        
        if len(transcript.get_cds_sequence()) % 3 != 0:
            raise ValueError("anomalous_coding_sequence")
        
        # ignore mitochondrial genes, since mitochondiral mutation rates differ
        # from autosomal and allosomal mutation rates
        if transcript.get_chrom() == "MT":
            continue
        
        if combined_transcript is None:
            sites = SiteRates(transcript, mut_dict)
            combined_transcript = transcript
        else:
            sites = SiteRates(transcript, mut_dict, masked_sites=combined_transcript)
            combined_transcript += transcript
        
        missense_rates = sites["missense"]
        nonsense_rates = sites["nonsense"]
        splice_lof_rates = sites["splice_lof"]
        splice_region_rates = sites["splice_region"]
        synonymous_rates = sites["synonymous"]
        
        # if any sites have been sampled in the transcript, then add the
        # cumulative probability from those sites to the approporiate
        # mutation rate. Sometimes we won't have any sites for a transcript, as
        # all the sites will have been captured in previous transcripts.
        missense += missense_rates.get_summed_rate()
        nonsense += nonsense_rates.get_summed_rate()
        splice_lof += splice_lof_rates.get_summed_rate()
        splice_region += splice_region_rates.get_summed_rate()
        synonymous += synonymous_rates.get_summed_rate()
    
    chrom = combined_transcript.get_chrom()
    length = "NA"
    if combined_transcript is not None:
        length = combined_transcript.get_coding_distance(\
            combined_transcript.get_cds_start(), combined_transcript.get_cds_end())
    
    return (chrom, length, missense, nonsense, splice_lof, splice_region, synonymous)

def log_transform(values):
    """ log transform a numeric value, unless it is zero, or negative
    """
    
    transformed = []
    for value in values:
        try:
            value = math.log10(value)
        except ValueError:
            value = "NA"
        transformed.append(value)
    
    return transformed

def include_indel_rates(path):
    """ add per-gene indel mutation rates to the output file
    
    We estimate the indel mutation rate as per Nature Genetics 46:944-950
    (2014) doi:10.1038/ng.3050, in which the overall (ie summed across all
    genes) indel mutation rate is estimated as 1.25 x the overall nonsense
    rate. Given this, the indel mutation rate is portioned out to each gene,
    according to the proportion of each gene's CDS length to the length of all
    CDS regions from all genes.
    
    Note that this is an approximation, and doesn't allow for divergence due to
    the base compostion of the CDS.
    
    Args:
        path: path to the output mutation rates (for the nonsense, missense etc)
    """
    
    # run through the file of rates to find the overall nonsense mutation rate,
    # and the total length of CDS regions in the file.
    nonsense_sum = 0
    length_sum = 0
    lines = []
    with open(path) as handle:
        nonsense_pos = None
        length_pos = None
        for line in handle:
            lines.append(line)
            if line.startswith("transcript_id"):
                nonsense_pos = line.split("\t").index("nonsense_rate")
                length_pos = line.split("\t").index("length")
                continue
            
            line = line.strip().split("\t")
            nonsense_sum += 10**float(line[nonsense_pos])
            length_sum += int(line[length_pos])
    
    # add the frameshift rates to each line in turn, while writing the output
    # back to the output path
    frameshift_sum = nonsense_sum * 1.25
    with open(path, "w") as handle:
        for line in lines:
            line = line.strip().split("\t")
            if line[0] == "transcript_id":
                line.append("frameshift_rate")
            else:
                # estimate the frameshift rate for the gene
                frameshift_rate = (float(line[2])/length_sum) * frameshift_sum
                frameshift_rate = math.log10(frameshift_rate)
                line.append(str(frameshift_rate))
            
            line = "\t".join(line) +"\n"
            handle.write(line)

def main():
    
    input_transcripts, input_genes, output_file, rates_file, cache_dir, \
        genome_build = get_options()
    
    # load all the data
    ensembl = EnsemblRequest(cache_dir, genome_build)
    mut_dict = load_mutation_rates(rates_file)
    
    if input_transcripts is not None:
        transcripts = load_transcripts(input_transcripts)
    else:
        transcripts = load_genes(input_genes)
    
    output = open(output_file, "w")
    output.write("transcript_id\tchrom\tlength\tmissense_rate\tnonsense_rate\t"
        "splice_lof_rate\tsplice_region_rate\tsynonymous_rate\n")
    
    for gene_id in sorted(transcripts):
        print(gene_id)
        try:
            rates = get_mutation_rates(gene_id, transcripts, mut_dict, ensembl)
            
            chrom = rates[0]
            length = rates[1]
            rates = rates[2:]
            # log transform rates, for consistency with Samocha et al.
            line = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\n".format(gene_id, \
                chrom, length, *log_transform(rates))
        except ValueError as error:
            line = "{0}\t{1}\n".format(gene_id, error)
        except KeyError as error:
            # ignore genes with odd genomic sequence eg ENST00000436041 in GRCh37
            continue
        
        output.write(line)
    
    output.close()
    
    include_indel_rates(output_file)

if __name__ == '__main__':
    main()



"""
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import unittest
import tempfile
import random
import os
import sys
import shutil
import zlib
from datetime import datetime, timedelta

from denovonear.ensembl_cache import EnsemblCache

IS_PYTHON2 = sys.version_info[0] == 2
IS_PYTHON3 = sys.version_info[0] == 3

class TestEnsemblCachePy(unittest.TestCase):
    """ unit test the EnsemblCache class
    """
    
    def setUp(self):
        """ construct an EnsemblCache object for unit tests
        """
        
        # set up a random string to use as a directory name
        hash_string = None
        while hash_string is None:
            string = "%8x" % random.getrandbits(64)
            string = string.strip()
            if not self.is_number(string):
                hash_string = string
        
        # make a temp directory for the cache file
        self.temp_dir = os.path.join(tempfile.gettempdir(), hash_string)
        os.mkdir(self.temp_dir)
        
        # construct the cache object
        self.cache = EnsemblCache(self.temp_dir, "grch37")
        self.cache.set_ensembl_api_version("3.0.0")
    
    def is_number(self, string):
        """ checks if a string can be converted to a number
        """
        
        try:
            float(string)
            return True
        except ValueError:
            return False
    
    def tearDown(self):
        """ remove the temp directory (and contents) containing the sql cache
        """
        
        shutil.rmtree(self.temp_dir)
    
    def test_retrieve_data(self):
        """ test that retrieve_data() works correctly
        """
        
        # check that we raise an error if the data hasn't been set already
        with self.assertRaises(AttributeError):
            self.cache.retrieve_data()
        
        self.cache.data = "aaa"
        self.assertEqual(self.cache.retrieve_data(), "aaa")
    
    def test_get_key_from_url(self):
        """ test that get_key_from_url() works correctly
        """
        
        self.assertEqual(self.cache.get_key_from_url("http://beta.rest.ensembl.org/info/rest"), "info.rest")
        self.assertEqual(self.cache.get_key_from_url("http://beta.rest.ensembl.org/xrefs/symbol/homo_sapiens/ABO"), "xrefs.symbol.homo_sapiens.ABO")
        self.assertEqual(self.cache.get_key_from_url("http://beta.rest.ensembl.org/sequence/id/ENST00000378520?type=protein"), "sequence.id.ENST00000378520.protein")
        self.assertEqual(self.cache.get_key_from_url("http://beta.rest.ensembl.org/feature/id/ENSG00000175164?feature=transcript"), "feature.id.ENSG00000175164.transcript")
        self.assertEqual(self.cache.get_key_from_url("http://beta.rest.ensembl.org/sequence/id/ENST00000538324?type=genomic;expand_3prime=10;expand_5prime=10"), "sequence.id.ENST00000538324.genomic")
        self.assertEqual(self.cache.get_key_from_url("http://beta.rest.ensembl.org/sequence/id/ENST00000538324?type=cds"), "sequence.id.ENST00000538324.cds")
        self.assertEqual(self.cache.get_key_from_url("http://beta.rest.ensembl.org/feature/id/ENST00000538324?feature=exon"), "feature.id.ENST00000538324.exon")
        self.assertEqual(self.cache.get_key_from_url("http://beta.rest.ensembl.org/vep/human/id/rs3887873/consequences?"), "vep.human.id.rs3887873.consequences")
        self.assertEqual(self.cache.get_key_from_url("http://beta.rest.ensembl.org/vep/human/9:22125503-22125502:1/C/consequences?"), "vep.human.9_22125503-22125502_1.C.consequences")
    
    def test_check_if_data_in_cache(self):
        """ test that check_if_data_in_cache() works correctly
        """
        
        # set up the data to go in the database
        current_date = datetime.strftime(datetime.today(), "%Y-%m-%d")
        genome_build = "grch37"
        url = "http://beta.rest.ensembl.org/feature/id/temp1?feature=exon"
        key = self.cache.get_key_from_url(url)
        api_version = self.cache.api_version
        string = "temp_data"
        if IS_PYTHON3:
            string = string.encode("utf-8")
        temp_data = zlib.compress(string)
        
        if IS_PYTHON2:
            temp_data = buffer(temp_data)
        
        # check that the data not in the database returns False
        self.assertFalse(self.cache.check_if_data_in_cache(url))
        
        # insert the data in the database
        values = (key, genome_build, current_date, api_version, temp_data)
        self.cache.c.execute("INSERT into ensembl VALUES (?, ?, ?, ?, ?)", values)
        
        # check that the data in the database returns True
        self.assertTrue(self.cache.check_if_data_in_cache(url))
        
        # check that the cache's data object is set correctly if the row is in the database
        if IS_PYTHON2:
            self.assertEqual(self.cache.data, zlib.decompress(temp_data))
        elif IS_PYTHON3:
            self.assertEqual(self.cache.data, zlib.decompress(temp_data).decode("utf-8"))
        
        # check that obsolete data returns False
        old_date = datetime.strftime(datetime.today() - timedelta(days=181), "%Y-%m-%d")
        values = (key, genome_build, old_date, api_version, temp_data)
        self.cache.c.execute("REPLACE into ensembl VALUES (?, ?, ?, ?, ?)", values)
        self.assertFalse(self.cache.check_if_data_in_cache(url))
        
        # check that data from an old API version returns False
        old_api_version = "1.0.0"
        values = (key, genome_build, current_date, old_api_version, temp_data)
        self.cache.c.execute("REPLACE into ensembl VALUES (?, ?, ?, ?, ?)", values)
        self.assertFalse(self.cache.check_if_data_in_cache(url))
    
    def test_cache_url_data(self):
        """ test that cache_url_data works correctly
        """
        
        # set up the data to go in the database
        current_date = datetime.strftime(datetime.today(), "%Y-%m-%d")
        url = "http://beta.rest.ensembl.org/feature/id/temp1?feature=exon"
        key = self.cache.get_key_from_url(url)
        api_version = self.cache.api_version
        temp_data = "temp_data"
        
        # check that the data is not in before we insert it
        self.assertFalse(self.cache.check_if_data_in_cache(url))
        
        # insert the data, then check that it has gone in
        self.cache.cache_url_data(url, temp_data)
        self.assertTrue(self.cache.check_if_data_in_cache(url))
        
        # check that we can't cache the ensembl version string (so we have to
        # check this each time the script runs, rather than pulling from the
        # cache)
        url = "http://beta.rest.ensembl.org/info/rest"
        self.cache.cache_url_data(url, temp_data)
        self.assertFalse(self.cache.check_if_data_in_cache(url))



""" class to plot transcripts.
"""

from __future__ import division


class TranscriptPlot(object):
    """ class to plot transcripts.
    """
    
    def plot_transcript(self):
        """ plots a transcript
        """
        
        # increment the y_offset position, so as to avoid overplotting between
        # gene, transcript and protein diagrams
        self.y_offset += self.box_height * 3
        
        cds_start = self.transcript.get_cds_start()
        cds_end = self.transcript.get_cds_end()
        length = (self.transcript.get_coding_distance(cds_start, cds_end) + 1) / self.size
        
        # give a label for the gene
        x_pos = 0
        self.add_text(x_pos, self.transcript.get_name(), y_adjust=self.box_height*1.5)
        
        for exon in self.transcript.get_cds():
            start, end = exon['start'], exon['end']
            if self.transcript.get_strand() == "+":
                x_pos = self.transcript.get_coding_distance(cds_start, start) / length
                width = (self.transcript.get_coding_distance(start, end) + 1) / length
            else:
                x_pos = self.transcript.get_coding_distance(cds_start, end) / length
                width = (self.transcript.get_coding_distance(start, end) + 1) / length
            
            self.add_box(x_pos, width, fillcolor="green")
        
        # and plot the de novo positions
        for x in self.de_novos:
            cds = self.transcript.chrom_pos_to_cds(self.de_novos[x]["start_pos"])
            x_pos = cds['pos']/length
            width = max(1/length, self.size/1000)
            
            self.de_novos[x]["coordinate"] = (x_pos, width)
            
        self.add_de_novos(self.de_novos)
        
        # and include the gene symbol and amino acid length on the domain plot
        text = "{} bp".format(self.transcript.get_coding_distance(cds_start, cds_end) + 1)
        self.add_text(self.size, text, y_adjust=self.box_height*1.5, horizontalalignment="right")
        




# try interpro for domain information

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys
import json
import time
import logging
import platform

IS_PYTHON2 = sys.version_info[0] == 2
IS_PYTHON3 = sys.version_info[0] == 3

# load python version specific url opener library
if IS_PYTHON2:
    import urllib2 as request
    from urllib import urlencode
elif IS_PYTHON3:
    import urllib.request as request
    from urllib.parse import urlencode
else:
    raise ValueError("unknown python version")

from denovonear.gene_plot.interpro_parser import InterProParser

try:
    import xml.etree.cElementTree as et
except ImportError:
    import xml.etree.ElementTree as et

logging.basicConfig(filename='interpro_requests.log',level=logging.INFO)

class InterproRequest(InterProParser):
    """ Uses the InterPro REST API to obtain domain information from InterPro.
    """
    
    rate_limit = 0.067
    job_limit = 30
    title = "DomainPlotter"
    
    def __init__(self, email_address):
        """ obtain the sequence for a transcript from ensembl
        """
        
        self.prior_time = time.time() - 1
        self.submitted_jobs = 0
        self.results = []
        
        self.email = email_address
        self.server = "http://www.ebi.ac.uk/Tools/services/rest/iprscan5"
    
    def getUserAgent(self):
        """ User-agent for request (see RFC2616).
        """
        
        # Agent string for urllib2 library.
        urllib_agent = "Python-urllib {0}".format(request.__version__)
        clientVersion = '0'
        
        # Prepend client specific agent string.
        user_agent = "Interpro REST API ver{0}: Python {1} on {2} with {3}".format(
            clientVersion,
            platform.python_version(), platform.system(),
            urllib_agent
        )
        
        return user_agent
        
    def open_url(self, url, headers, post_data=False):
        """ open url with python libraries
        """
        
        req = request.Request(url, headers=headers)
        
        try:
            if post_data == False:
                handler = request.urlopen(req)
            else:
                data = urlencode(post_data)
                if IS_PYTHON3:
                    data = data.encode("utf-8")
                handler = request.urlopen(req, data)
        except Exception as e:
            handler = e
        
        status_code = handler.getcode()
        response = handler.read()
        if IS_PYTHON3:
            response = response.decode("utf-8")
        
        # parse the headers into a key, value dictionary
        headers = {}
        for key, value in zip(handler.headers.keys(), handler.headers.values()):
            headers[key.lower()] = value
        
        return response, status_code, headers
    
    def api_request(self, ext, headers, post_data=False):
        """ obtain sequence via the EBI InterPro REST API
        """
        
        self.request_attempts += 1
        if self.request_attempts > 5:
            raise ValueError("too many attempts, figure out why its failing")
        
        self.rate_limit_requests()
        response, status_code, requested_headers = self.open_url(self.server + ext, headers=headers, post_data=post_data)
        logging.info("{0}\t{1}\t{2}".format(time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime()), status_code, self.server + ext))
        
        if status_code != 200:
            raise ValueError("Invalid response: " + str(status_code)\
                + ".\nSubmitted URL was: " + self.server + ext + "\nheaders: " \
                + str(requested_headers) + "\nresponse: " + response)
        
        return response
    
    def rate_limit_requests(self):
        """ limit api requests to one per 0.335 s
        """
        
        sleeping = True
        while sleeping:
            if (time.time() - self.prior_time) > self.rate_limit:
                self.prior_time = time.time()
                sleeping = False
                
            time.sleep(0.01)
    
    def poll_until_job_finished(self, job_id):
        """ given a submitted job, periodically query until the job has finished
        """
        
        print("server is analysing: {0}".format(job_id))
        
        job_status = "PENDING"
        while job_status != "FINISHED":
            self.request_attempts = 0
            time.sleep(30)
            job_status = self.check_job_status(job_id)
            print(job_status)
            
            # quit out on various errors described at:
            # http://www.ebi.ac.uk/Tools/webservices/services/pfa/iprscan5_rest
            if job_status == "ERROR" or job_status == "FAILURE" or job_status == "NOT_FOUND":
                raise ValueError("submitted job has gone wrong: " + job_id)
            
        return self.get_result(job_id)
    
    def submit_interpro_run(self, protein_sequence):
        """ submit interpro run, to predict domains in a protein sequence
        """
        
        self.request_attempts = 0
        
        post_data = {}
        post_data["email"] = self.email
        post_data["title"] = self.title
        post_data["sequence"] = protein_sequence.rstrip() + "\n"
        
        header = {"User-Agent": self.getUserAgent()}
        
        job_id = self.api_request("/run/", header, post_data)
        self.job_id = job_id
        self.submitted_jobs += 1
        
        return self.poll_until_job_finished(job_id)
    
    def check_job_status(self, job_id):
        
        self.request_attempts = 0
        header = {"User-Agent": self.getUserAgent()}
        status = self.api_request("/status/{}".format(job_id), header)
        
        return status
    
    def get_result_types(self, job_id):
        
        self.request_attempts = 0
        header = {"User-Agent": self.getUserAgent()}
        result_types = self.api_request("/resulttypes/{}".format(job_id), header)
        
        root = et.fromstring(result_types)
        
        results = []
        for child in root:
            entry = {}
            entry["descriptor"] = child.find("description").text
            entry["identifier"] = child.find("identifier").text
            entry["file_suffix"] = child.find("fileSuffix").text
            entry["label"] = child.find("label").text
            entry["media_type"] = child.find("mediaType").text
            
            results.append(entry)
        
        return results
    
    def get_result(self, job_id):
        """ obtain the domain predictions for a InterPro REST API job ID
        """
        
        self.request_attempts = 0
        header = {"User-Agent": self.getUserAgent()}
        api_response = self.api_request("/result/{}/tsv".format(job_id), header)
        
        domains = self.parse_interpro_results(api_response)
        domains = self.standardise_names(domains)
        
        return domains



""" unit test the  class"""

import os
import unittest
import tempfile
import shutil
import math

from denovonear.ensembl_requester import EnsemblRequest
from denovonear.load_mutation_rates import load_mutation_rates
from denovonear.cluster_test import fishers_method, cluster_de_novos, combine_p_values

class TestClusterTestPy(unittest.TestCase):
    """ unit test the genic test
    """
    
    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.mkdtemp()
        cls.ensembl = EnsemblRequest(cls.temp_dir, 'grch37')
        cls.mut_dict = load_mutation_rates()
    
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.temp_dir)
    
    def test_cluster_de_novos(self):
        """ check that cluster_de_novos() runs correctly
        """
        
        symbol = 'PPP2R5D'
        de_novos = {'missense': [42975003, 42975003, 42975003, 42975013],
            'nonsense': []}
        
        p_values = cluster_de_novos(symbol, de_novos, 1000000,
            self.ensembl, self.mut_dict)
        
        self.assertAlmostEqual(p_values['miss_prob'], 3e-06, delta=3e-6)
        self.assertEqual(p_values['miss_dist'], '2.3')
        self.assertTrue(math.isnan(p_values['nons_prob']))
        self.assertEqual(p_values['nons_dist'], 'nan')
        
        # TODO: add a test with de novos that occur on different transcripts
    
    def test_fishers_method(self):
        """ check that fishers combined method calculates correctly
        """
        
        p_values = [0.01, 0.001]
        self.assertAlmostEqual(fishers_method(p_values), 0.0001251292546)
        
        # correct for single value. This is off by a minuscule fraction due to
        # float division
        p_values = [0.01]
        self.assertAlmostEqual(fishers_method(p_values), 0.01)
        
        # correct for list with NA value
        p_values = [0.01, float('nan')]
        self.assertAlmostEqual(fishers_method(p_values), 0.01)
        
        # correct for list with NA value
        p_values = [0.01, 0.001, float('nan')]
        self.assertAlmostEqual(fishers_method(p_values), 0.0001251292546)
        
        # correct without any p-values
        p_values = [float('nan'), float('nan')]
        self.assertTrue(math.isnan(fishers_method(p_values)))
        
        # raise error for list with zero
        with self.assertRaises(ValueError):
            p_values = [0, 0.01]
            fishers_method(p_values)
    
    def test_combine_p_values(self):
        """ check that combine_p_values() works correctly
        """
        
        probs = {"miss_prob": [0.01, 0.001], "nons_prob": [0.01, 0.01]}
        data = combine_p_values(probs)
        
        self.assertAlmostEqual(data['miss_prob'], 0.0001251292546)
        self.assertAlmostEqual(data['nons_prob'], 0.0010210340371)




import sys
import os
import io
from setuptools import setup
from distutils.core import Extension
from Cython.Build import cythonize

EXTRA_COMPILE_ARGS = ["-std=c++11"]

if sys.platform == "darwin":
    EXTRA_COMPILE_ARGS += ["-stdlib=libc++"]

weights = cythonize([
    Extension("denovonear.weights",
        extra_compile_args=EXTRA_COMPILE_ARGS,
        sources=["denovonear/weights.pyx",
            "src/weighted_choice.cpp",
            "src/simulate.cpp"],
        include_dirs=["src/"],
        language="c++"),
    Extension("denovonear.transcript",
        extra_compile_args=EXTRA_COMPILE_ARGS,
        sources=[
            "denovonear/transcript.pyx",
            "src/tx.cpp"],
        include_dirs=["src/"],
        language="c++"),
    Extension("denovonear.site_specific_rates",
        extra_compile_args=EXTRA_COMPILE_ARGS,
        sources=[
            "denovonear/site_specific_rates.pyx",
            "src/weighted_choice.cpp",
            "src/tx.cpp",
            "src/site_rates.cpp"],
        include_dirs=["src/"],
        language="c++"),
    ])

setup (name="denovonear",
        description='Package to examine de novo clustering',
        long_description=io.open('README.rst', encoding='utf-8').read(),
        version="0.5.3",
        author="Jeremy McRae",
        author_email="jeremy.mcrae@sanger.ac.uk",
        license="MIT",
        url='https://github.com/jeremymcrae/denovonear',
        packages=["denovonear", "denovonear.gene_plot"],
        install_requires=['scipy >= 0.9.0',
                          'cairocffi >= 0.7.2',
                          'webcolors >= 1.5',
                          'cython >= 0.19.0'
        ],
        package_data={"denovonear": ['data/rates.txt', 'weights.pxd']},
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "License :: OSI Approved :: MIT License",
        ],
        ext_modules=weights,
        test_suite="tests")



""" caches sequence information requested from ensembl, so we don't have to
repeatedly re-request information from the REST API if we have done so recently.
"""

import os
import sqlite3
import sys
import time
import zlib
from datetime import datetime

IS_PYTHON2 = sys.version_info[0] == 2
IS_PYTHON3 = sys.version_info[0] == 3

class EnsemblCache(object):
    """ Instead of repeatedly re-acquiring data from Ensembl each run, cache
    the requested data for faster retrieval
    """
    
    def __init__(self, cache_folder, genome_build):
        """ initialise the class with the local cache folder
        
        Args:
            cache_folder: path to the cache
        """
        
        self.cache_folder = cache_folder
        self.cache_path = os.path.join(self.cache_folder, "ensembl_cache.db")
        
        self.genome_build = genome_build
        self.today = datetime.today()
        
        self.connect_to_database()
    
    def connect_to_database(self):
        """ connect to a sqlite database for the cached data (creates a new
        database if it doesn't already exist).
        """
        
        # create a folder for the cache database (this will raise an error
        # if we cannot create the folder)
        if not os.path.exists(self.cache_folder):
            os.mkdir(self.cache_folder)
        
        # generate a database with tables if it doesn't already exist
        if not os.path.exists(self.cache_path):
            conn = sqlite3.connect(self.cache_path)
            c = conn.cursor()
            try:
                c.execute(u"CREATE TABLE ensembl (key text PRIMARY KEY, genome_build text, cache_date text, api_version text, data blob)")
                conn.commit()
                c.close()
            except sqlite3.OperationalError:
                # occurs when multiple processes simultaneously try to create the
                # database
                
                # briefly sleep, so the process creating the database has time to
                # construct it
                c.close()
                time.sleep(5)
            
        self.conn = sqlite3.connect(self.cache_path)
        self.conn.row_factory = sqlite3.Row
        self.c = self.conn.cursor()
    
    def set_ensembl_api_version(self, version):
        """ set the ensembl API version, so we can check for obsolete data
        
        Args:
            version: Ensembl API version string eg "2.0.0"
        """
        
        self.api_version = version
    
    def check_if_data_in_cache(self, url):
        """ checks if the data for a url has already been stored in the cache
        
        Makes sure that the data is not out of date, by only allowing data that
        is less than six months old, and produced by the same Ensembl REST API
        version.
        
        Args:
            url: URL for the Ensembl REST service
        
        Returns:
            True/False for having the data in the cache
        """
        
        key = self.get_key_from_url(url)
        
        self.c.execute("SELECT * FROM ensembl WHERE key =? AND genome_build=?", (key, self.genome_build))
        row = self.c.fetchone()
        
        # if the data has been cached, check that it is not out of date, and
        # the data was generated from the same Ensembl API version
        if row is not None:
            api_version = row["api_version"]
            
            # by default extract the data, so we don't have to repeat the
            # sql query later.
            self.data = zlib.decompress(row["data"])
            if IS_PYTHON3:
                self.data = self.data.decode("utf-8")
            
            cache_date = datetime.strptime(row["cache_date"], "%Y-%m-%d")
            diff = self.today - cache_date
            
            if diff.days < 180 and self.api_version == api_version:
                return True
        
        return False
    
    def retrieve_data(self):
        """retrieves data for a URL (checked earlier that this exists)
        
        The function is only used if the data exists in the cache. In order to
        check if the data is in the cache, we selected database rows, and
        parsed the contents (to check if the data is obsolete). Rather than
        reselecting the data, just use the data loaded earlier.
        
        Returns:
            the data object for the URL (which will be parsed elsewhere)
        """
        
        return self.data
    
    def cache_url_data(self, url, data):
        """ cache the data retrieved from ensembl
        
        Args:
            url: URL for the Ensembl REST service
            data: response data from Ensembl
        """
        
        key = self.get_key_from_url(url)
        
        # don't cache the ensembl version check
        if key == "info.rest":
            return
        
        current_date = datetime.strftime(self.today, "%Y-%m-%d")
        
        # python3 zlib requires encoded strings
        if IS_PYTHON3:
            data = data.encode("utf-8")
        
        data = zlib.compress(data)
        
        # python2 sqlite3 can't write "8-bit bytestrings", but it can handle
        # buffer versions of the bytestrings
        if IS_PYTHON2:
            data = buffer(data)
        
        t = (key, self.genome_build, current_date, self.api_version, data)
        
        self.c.execute("INSERT OR REPLACE INTO ensembl (key, genome_build, cache_date, api_version, data) VALUES (?,?,?,?,?)", t)
        self.conn.commit()
    
    def get_key_from_url(self, url):
        """ parses the url into a list of folder locations
        
        Args:
            url: URL for the Ensembl REST service
        
        Returns:
            a parsed unique database key for the URLs data
        """
        
        key = url.split("/")[3:]
        
        # fix the final bit of the url, which can have additional requirements,
        # none of which are necessary for uniquely defining the data
        key[-1] = key[-1].split(";")[0]
        
        # convert "feature=transcript" to "transcript" or "type=protein" to
        # "protein"
        final = key[-1].split("?")
        if "=" in final[-1]:
            final[-1] = final[-1].split("=")[1]
        
        key = key[:-1] + final
        
        # replace characters not tolerated in keys
        for pos in range(len(key)):
            key[pos] = key[pos].replace(":", "_")
        
        # if the url ended with "?", then the final list element will be "",
        # which we should remove
        if key[-1] == "":
            key = key[:-1]
        
        key = ".".join(key)
        
        return key



""" counts high-frequency LOF variants in genes in human genome.

Uses 1000 Genomes variation data to find variants in genes, and estimate the
minor allele frequency in continental populations.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import argparse

from denovonear.load_gene import get_deprecated_gene_ids, get_transcript_lengths, \
    construct_gene_object, get_de_novos_in_transcript, load_gene
from denovonear.thousand_genomes.ensembl_consequences import EnsemblWithVariants
from denovonear.thousand_genomes.variants import Extract1000Genomes


CONSEQUENCES = {"lof": set(["transcript_ablation","splice_donor_variant", \
    "splice_acceptor_variant", "frameshift_variant", "stop_gained", \
    "coding_sequence_variant"]), \
    "missense": set(["initiator_codon_variant", "inframe_insertion",
    "inframe_deletion", "missense_variant", "transcript_amplification",
    "stop_lost", "splice_region_variant"]), \
    "synonymous": set(["synonymous_variant"])}

def get_options():
    """
    """
    
    parser = argparse.ArgumentParser(description="Counts common LoF variants in genes.")
    parser.add_argument("--out", dest="output", required=True, help="output \
        filename")
    parser.add_argument("--min-freq", dest="min_freq", default=0.05, help="Minimum accepted minor genotype frequency")
    parser.add_argument("--max-freq", dest="max_freq", default=0.6, help="Maximum accepted minor genotype frequency")
    parser.add_argument("--hgnc", dest="hgnc_filename", default="/nfs/users/nfs_j/jm33/gel_annotation_exercise/data/ensembl.gene_symbols_and_chrom_positions.txt", help="file listing genes to analyse")
    parser.add_argument("--cache-folder", dest="cache_folder", \
        default=os.path.join(os.path.dirname(__file__), "cache"), help="folder \
        to cache Ensembl data into (defaults to clustering code directory)")
    parser.add_argument("--genome-build", dest="genome_build", choices=["grch37",
        "GRCh37", "grch38", "GRCh38"], default="grch37", help="Genome build "+ \
        "that the de novo coordinates are based on (GrCh37 or GRCh38")
    parser.add_argument("--sum-frequency", default=False, action="store_true")
    
    args = parser.parse_args()
    
    return args.output, float(args.min_freq), float(args.max_freq), \
        args.hgnc_filename, args.cache_folder, args.genome_build, args.sum_frequency


def load_hgnc_symbols(filename):
    """ load a file containing all the current HGNC symbols
    """
    
    genes = set([])
    with open(filename, "r") as f:
        for line in f:
            line = line.split("\t")
            hgnc = line[0]
            
            # ignore blank lines
            if hgnc == "":
                continue
            
            genes.add(hgnc)
    
    return sorted(genes)

def get_already_loaded_genes(filename):
    """find the genes which have already been assessed, so we can resume from
    where we left off (if the code or system breaks)
    """
    
    prior_genes = set([])
    if os.path.exists(filename):
        current_genes = open(filename, "r")
        for line in current_genes:
            line = line.strip().split("\t")
            gene = line[0]
            prior_genes.add(gene)
        current_genes.close()
    
    return prior_genes

def main():
    
    output_filename, min_freq, max_freq, hgnc_filename, cache_dir, genome_build, check_frequency = get_options()
    
    thousand_genomes = Extract1000Genomes("/lustre/scratch113/teams/hurles/users/jm33/1000genomes/", frequency="genotype")
    ensembl = EnsemblWithVariants(cache_dir, genome_build)
    hgnc_symbols = load_hgnc_symbols(hgnc_filename)
    
    # find the genes which have already been assessed, so we can resume from
    # where we left off (if the code or system breaks)
    prior_genes = get_already_loaded_genes(output_filename)
    
    output = open(output_filename, "a")
    
    for hgnc in hgnc_symbols:
        if hgnc in prior_genes:
            continue
        prior_genes.add(hgnc)
        
        # load the transcript ID, exon positions and sequence. If we raise an
        # error, just move to the next gene (eg ABO, which tries to load a
        # transcript without protein coding sequence, rather than a transcript
        # on a patch contig).
        try:
            gene = load_gene(ensembl, hgnc)
        except (IndexError, ValueError) as e:
            continue
        
        thousand_genomes.set_gene(gene)
        if not check_frequency:
            func = thousand_genomes.filter_variants(min_maf=min_freq, max_maf=max_freq, ignore_indels=False)
            output.write("{0}\t{1}\n".format(hgnc, len(func["lof"])))
        else:
            (samples, n) = thousand_genomes.find_samples_with_disrupted_alleles_in_gene()
            try:
                lof_freq = len(samples["lof"])/n
                missense_freq = len(samples["missense"])/n
            except ZeroDivisionError:
                lof_freq, missense_freq = "NA", "NA"
            output.write("{0}\t{1}\t{2}\n".format(hgnc, lof_freq, missense_freq))
        # chrom = gene.chrom
        # transcript_id = gene.name
        # for category in func:
        #     for (var, ref, alt) in func[category]:
        #         if var.id != ".":
        #             consequence = ensembl.get_variant_by_id(var.id, transcript_id, chrom, var.pos, ref, alt, gene)
                
        #         if consequence not in CONSEQUENCES[category]:
        #             exon_start, exon_end = gene.find_closest_exon(var.pos)
        #             boundary_dist = min(abs(exon_start - var.pos), abs(exon_end - var.pos))
        #             print(var.id, var.pos, category, consequence, boundary_dist)
        
        # break
        

if __name__ == "__main__":
    main()



""" Obtains genomic sequence, transcript IDs, gene, exon and CDS coordinates for
transcripts from Ensembl.
"""

from __future__ import print_function

import sys
import json
import time
import logging

IS_PYTHON3 = sys.version_info[0] == 3
 
if IS_PYTHON3:
    import urllib.request as request
    from urllib.error import HTTPError
else:
    import urllib2 as request
    from urllib2 import HTTPError

from denovonear.ensembl_cache import EnsemblCache

logging.basicConfig(filename='ensembl_requests.log', level=logging.WARNING)

class EnsemblRequest(object):
    """ Uses the Ensembl REST API to obtain gene information from Ensembl.
    
    Can find:
         - gene IDs for a HGNC symbol
         - transcript IDs for a gene ID
         - exon coordinates for an ensembl transcript ID
         - CDS coordinates for an ensembl transcript ID
         - transcript and genomic DNA sequences for an ensembl transcript ID
    """
    
    def __init__(self, cache_folder, genome_build):
        """ obtain the sequence for a transcript from ensembl
        
        Args:
            cache_folder: path to folder for caching data requested from Ensembl
            genome_build: string indicating the genome build ("grch37" or "grch38")
        """
        
        self.cache = EnsemblCache(cache_folder, genome_build)
        
        self.prior_time = time.time() - 1
        self.rate_limit = 0.067
        
        server_dict = {"grch37": "grch37.", "grch38": ""}
        
        self.server = "http://{}rest.ensembl.org".format(server_dict[genome_build])
        
        self.check_ensembl_api_version()
    
    def check_ensembl_api_version(self):
        """ check the ensembl api version matches a currently working version
        
        This function is included so when the api version changes, we notice the
        change, and we can manually check the responses for the new version.
        """
        
        self.request_attempts = 0
        headers = {"Content-Type": "application/json"}
        ext = "/info/rest"
        r = self.ensembl_request(ext, "version_check", headers)
        
        response = json.loads(r)
        release = response["release"].split(".")
        self.cache.set_ensembl_api_version(response["release"])
        
        major = release[0]
        minor = release[1]
        
        if major != "6":
            raise ValueError("check ensembl api version")
        
    def open_url(self, url, headers):
        """ open url with python libraries
        """
        
        req = request.Request(url, headers=headers)
        
        try:
            handler = request.urlopen(req)
        except HTTPError as error:
            # if we get a http error, we still process the status code, since a
            # later step deals with different status codes differently.
            handler = error
        except ConnectionResetError:
            # if we get a ConnectionResetError, assume something has gone wrong
            # with the server. Later code will wait before retrying.
            return '', 500, headers
        
        status_code = handler.getcode()
        response = handler.read()
        if IS_PYTHON3:
            response = response.decode("utf-8")
        
        # parse the headers into a key, value dictionary
        headers = dict(zip(map(str.lower, handler.headers.keys()), handler.headers.values()))
        
        return response, status_code, headers
    
    def ensembl_request(self, ext, sequence_id, headers):
        """ obtain sequence via the ensembl REST API
        """
        
        self.request_attempts += 1
        if self.request_attempts > 5:
            raise ValueError("too many attempts, figure out why its failing")
            
        if self.cache.check_if_data_in_cache(self.server + ext):
            return self.cache.retrieve_data()
        
        self.rate_limit_ensembl_requests()
        response, status_code, requested_headers = self.open_url(self.server + ext, headers=headers)
        logging.warning("{0}\t{1}\t{2}\t{3}".format(\
            time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime()), status_code, \
            sequence_id, self.server + ext))
        
        # we might end up passing too many simultaneous requests, or too many
        # requests per hour, just wait until the period is finished before
        # retrying
        if status_code == 429:
            if "retry-after" in requested_headers:
                time.sleep(float(requested_headers["retry-after"]))
            elif "x-ratelimit-reset" in requested_headers:
                time.sleep(int(requested_headers["x-ratelimit-reset"]))
            
            return self.ensembl_request(ext, sequence_id, headers)
        # retry after 30 seconds if we get service unavailable error
        elif status_code in [500, 503, 504]:
            time.sleep(30)
            return self.ensembl_request(ext, sequence_id, headers)
        elif status_code != 200:
            raise ValueError("Invalid Ensembl response: {0} for {1}.\nSubmitted URL was: {2}{3}\nheaders: {4}\nresponse: {5}".format(status_code, sequence_id, \
                    self.server, ext, requested_headers, response))
        
        # sometimes ensembl returns odd data. I don't know what it is, but the
        # json interpreter can't handle it. Rather than trying to catch it,
        # simply re-request the data
        if requested_headers["content-type"] == "application/json":
            try:
                json.loads(response)
            except ValueError:
                logging.warning("{0}\t{1}\t{2}\t{3}\t{4}".format(\
                    time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime()), \
                    status_code, sequence_id, self.server + ext, \
                    "cannot obtain json output"))
                return self.ensembl_request(ext, sequence_id, requested_headers)
        
        self.cache.cache_url_data(self.server + ext, response)
        
        return response
    
    def get_genes_for_hgnc_id(self, hgnc_symbol):
        """ obtain the ensembl gene IDs that correspond to a HGNC symbol
        """
        
        headers = {"Content-Type": "application/json"}
        
        # http://grch37.rest.ensembl.org/xrefs/symbol/homo_sapiens/KMT2A?content-type=application/json
        
        self.request_attempts = 0
        ext = "/xrefs/symbol/homo_sapiens/{0}".format(hgnc_symbol)
        r = self.ensembl_request(ext, hgnc_symbol, headers)
        
        genes = []
        for item in json.loads(r):
            if item["type"] == "gene":
                genes.append(item["id"])
        
        return genes
    
    def get_previous_symbol(self, hgnc_symbol):
        """ sometimes we get HGNC symbols that do not match the ensembl rest version
        that we are currently using. We can look for earlier HGNC symbols for
        the gene using the service at rest.genenames.org
        
        Args:
            hgnc_symbol: HGNC symbol for the gene (eg "MLL2")
        
        Returns:
            list of deprecated gene symbols (eg ["KMT2A"])
        """
        
        ensembl_server = self.server
        gene_names_server = "http://rest.genenames.org"
        
        self.server = gene_names_server
        headers = {"accept": "application/json"}
        ext = "/fetch/symbol/{0}".format(hgnc_symbol)
        try:
            r = self.ensembl_request(ext, hgnc_symbol, headers)
        finally:
            self.server = ensembl_server
        
        gene_json = json.loads(r)
        
        prev_gene = []
        docs = gene_json["response"]["docs"]
        
        # strip out any gene entries that have been invalidated
        docs = [ x for x in docs if x["status"] != "Entry Withdrawn"]
        
        if len(docs) == 0:
            pass
        elif len(docs) > 1:
            raise ValueError("{0} has more than one alternate symbol, which I haven't accounted for.".format(hgnc_symbol))
        elif "prev_symbol" in docs[0]:
            prev_gene = docs[0]["prev_symbol"]
        
        return prev_gene
        
    def get_transcript_ids_for_ensembl_gene_ids(self, gene_ids, hgnc_symbols):
        """ fetch the ensembl transcript IDs for a given ensembl gene ID.
        
        Args:
            gene_ids: list of Ensembl gene IDs for the gene
            hgnc_symbols: list of possible HGNC symbols for gene
        """
    
        chroms = {"1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", \
             "12", "13", "14", "15", "16", "17", "18", "19", "20", "21", "22", \
              "X", "Y"}
        
        headers = {"Content-Type": "application/json"}
        
        transcript_ids = []
        for gene_id in gene_ids:
            self.request_attempts = 0
            ext = "/overlap/id/{0}?feature=transcript".format(gene_id)
            r = self.ensembl_request(ext, gene_id, headers)
            
            for item in json.loads(r):
                # ignore non-coding transcripts
                if item["biotype"] not in ["protein_coding", "polymorphic_pseudogene"]:
                    continue
                
                # ignore transcripts not on the standard chromosomes
                # (non-default chroms fail to map the known de novo variants
                # to the gene location
                if item["Parent"] != gene_id or item["seq_region_name"] not in \
                        chroms or \
                        all([symbol not in item["external_name"] for symbol in hgnc_symbols]):
                    continue
                transcript_ids.append(item["id"])
        
        return transcript_ids
    
    def get_genomic_seq_for_transcript(self, transcript_id, expand):
        """ obtain the sequence for a transcript from ensembl
        """
        
        headers = {"Content-Type": "application/json"}
        
        self.request_attempts = 0
        ext = "/sequence/id/{0}?type=genomic;expand_3prime={1};expand_5prime={1}".format(transcript_id, expand)
        r = self.ensembl_request(ext, transcript_id, headers)
        
        gene = json.loads(r)
        
        seq = gene["seq"]
        seq_id = gene["id"]
        
        if seq_id != transcript_id:
            raise ValueError("ensembl gave the wrong transcript")
        
        desc = gene["desc"].split(":")
        chrom = desc[2]
        start = int(desc[3]) + expand
        end = int(desc[4]) - expand
        strand_temp = int(desc[5])
        
        strand = "+"
        if strand_temp == -1:
            strand = "-"
        
        return (chrom, start, end, strand, seq)
    
    def get_cds_seq_for_transcript(self, transcript_id):
        """ obtain the sequence for a transcript from ensembl
        """
        
        headers = {"Content-Type": "text/plain"}
        
        self.request_attempts = 0
        ext = "/sequence/id/{0}?type=cds".format(transcript_id)
        
        return self.ensembl_request(ext, transcript_id, headers)
    
    def get_protein_seq_for_transcript(self, transcript_id):
        """ obtain the sequence for a transcript from ensembl
        """
        
        headers = {"Content-Type": "text/plain"}
        
        self.request_attempts = 0
        ext = "/sequence/id/{0}?type=protein".format(transcript_id)
        
        return self.ensembl_request(ext, transcript_id, headers)
    
    def get_genomic_seq_for_region(self, chrom, start_pos, end_pos):
        """ obtain the sequence for a genomic region
        """
        
        headers = {"Content-Type": "text/plain"}
        
        self.request_attempts = 0
        ext = "/sequence/region/human/{0}:{1}..{2}:1".format(chrom, start_pos, end_pos)
        
        return self.ensembl_request(ext, "{0}:{1}-{2}".format(chrom, start_pos, end_pos), headers)
    
    def get_chrom_for_transcript(self, transcript_id, hgnc_id):
        """ obtain the sequence for a transcript from ensembl
        """
        
        headers = {"Content-Type": "application/json"}
        
        self.request_attempts = 0
        ext = "/overlap/id/{0}?feature=gene".format(transcript_id)
        r =  self.ensembl_request(ext, transcript_id, headers)
        
        for gene in json.loads(r):
            if gene["external_name"] == hgnc_id:
                return gene["seq_region_name"]
        
        return None
    
    def get_exon_ranges_for_transcript(self, transcript_id):
        """ obtain the sequence for a transcript from ensembl
        """
        
        headers = {"Content-Type": "application/json"}
        
        self.request_attempts = 0
        ext = "/overlap/id/{0}?feature=exon".format(transcript_id)
        r = self.ensembl_request(ext, transcript_id, headers)
        
        exon_ranges = []
        for exon in json.loads(r):
            if exon["Parent"] != transcript_id:
                continue
            
            start = exon["start"]
            end = exon["end"]
            
            exon_ranges.append((start, end))
        
        return exon_ranges
    
    def get_cds_ranges_for_transcript(self, transcript_id):
        """ obtain the sequence for a transcript from ensembl
        """
        
        headers = {"Content-Type": "application/json"}
        
        self.request_attempts = 0
        ext = "/overlap/id/{0}?feature=cds".format(transcript_id)
        r = self.ensembl_request(ext, transcript_id, headers)
        
        cds_ranges = []
        for cds_range in json.loads(r):
            if cds_range["Parent"] != transcript_id:
                continue
            
            start = cds_range["start"]
            end = cds_range["end"]
            
            cds_ranges.append((start, end))
        
        return cds_ranges
    
    def rate_limit_ensembl_requests(self):
        """ limit ensembl requests to one per 0.067 s
        """
        
        current_time = time.time()
        diff_time = current_time - self.prior_time
        
        # sleep until the current rate limit period is finished
        if diff_time < self.rate_limit:
            time.sleep(self.rate_limit - diff_time)
        
        # set the access time to now, for later checks
        self.prior_time = time.time()



    



""" class to plot genes.
"""

from __future__ import division


class GenomicPlot(object):
    """ class to plot exons of genes using genomic coordinates.
    """
    
    def plot_gene(self):
        """ plots the exons of a gene
        """
        
        # increment the y_offset position, so as to avoid overplotting between
        # gene, transcript and protein diagrams
        self.y_offset += self.box_height * 3
        
        min_pos = self.transcript.get_start()
        max_pos = self.transcript.get_end()
        strand = self.transcript.get_strand()
        length = (max_pos - min_pos) / self.size
        
        # plot the base strand
        x_pos = (self.transcript.get_start() - min_pos) / length
        width = (self.transcript.get_end() - self.transcript.get_start()) / length
        height = self.box_height/8
        y_adjust = self.box_height/2 - height/2
        self.add_box(x_pos, width, height=height, y_adjust=y_adjust, fillcolor="black")
        
        # give a label for the gene
        self.add_text(x_pos, self.transcript.get_name(), y_adjust=self.box_height*1.5)
        
        for exon in self.transcript.get_exons():
            self.plot_single_exon(exon['start'], exon['end'], length, min_pos)
        
        for x in self.de_novos:
            x_pos = (self.de_novos[x]["start_pos"] - min_pos)/length
            width = max(1/length, self.size/1000)
            self.de_novos[x]["coordinate"] = (x_pos, width)
        
        self.add_de_novos(self.de_novos)
    
    def mixed_coords(self, start, end, length, min_pos):
        """
        """
        
        cds_min = min(self.transcript.get_cds_start(), self.transcript.get_cds_end())
        cds_max = max(self.transcript.get_cds_start(), self.transcript.get_cds_end())
        strand = self.transcript.get_strand()
        
        x_pos_1 = (start - min_pos) / length
        width_1 = (end - start) / length
        
        if start <= cds_min <= end:
            midpoint = cds_min
            color_1 = "white"
            color_2 = "green"
        else:
            midpoint = cds_max
            color_1 = "green"
            color_2 = "white"
        
        x_pos_2 = (midpoint - min_pos) / length
        width_2 = (end - midpoint) / length
        
        return (x_pos_1, width_1, color_1), (x_pos_2, width_2, color_2)
        
    def plot_single_exon(self, start, end, length, min_pos):
        """ adds a rectangle to the plot
        
        NOTE: this currently doesn't allow for single exon genes where the
        single exon has 5' and 3' untranslated regions as well as a coding
        region within the exon.
        """
        
        cds_min = min(self.transcript.get_cds_start(), self.transcript.get_cds_end())
        cds_max = max(self.transcript.get_cds_start(), self.transcript.get_cds_end())
        
        x_pos = (start - min_pos) / length
        width = (end - start) / length
        
        if self.transcript.in_coding_region(start) and self.transcript.in_coding_region(end):
            self.add_box(x_pos, width, fillcolor="green")
        elif not self.transcript.in_coding_region(start) and not self.transcript.in_coding_region(end):
            self.add_box(x_pos, width, fillcolor="white")
        else:
            # exons with coding and untranslated regions, either utr first, or
            # utr second
            (x_pos_1, width_1, color_1), (x_pos_2, width_2, color_2) = self.mixed_coords(start, end, length, min_pos)
            
            self.add_box(x_pos_1, width_1, fillcolor=color_1)
            self.add_box(x_pos_2, width_2, fillcolor=color_2)



""" Script to investigate the probability of multiple mutations clustering
within a single gene.
"""

from __future__ import print_function, division, absolute_import

import os
import argparse

from denovonear.ensembl_requester import EnsemblRequest
from denovonear.load_mutation_rates import load_mutation_rates
from denovonear.load_de_novos import load_de_novos
from denovonear.cluster_test import cluster_de_novos

def get_options():
    """ get the command line switches
    """
    
    parser = argparse.ArgumentParser(description="Examines the proximity" \
        "clustering of de novo mutations in genes.")
    parser.add_argument("--in", dest="input", required=True, help="Path to" \
        "file listing known mutations in genes. See example file in data folder" \
        "for format.")
    parser.add_argument("--out", dest="output", required=True, help="output \
        filename")
    parser.add_argument("--rates",
        help="Path to file containing sequence context-based mutation rates.")
    parser.add_argument("--deprecated-genes", dest="deprecated_genes_path",
        help="deprecated gene IDs filename")
    parser.add_argument("--genome-build", dest="genome_build", choices=["grch37",
        "GRCh37", "grch38", "GRCh38"], default="grch37", help="Genome build " \
        "that the de novo coordinates are based on (GrCh37 or GRCh38")
    parser.add_argument("--cache-folder", dest="cache_dir",
        default=os.path.join(os.path.dirname(__file__), "cache"), help="folder" \
        "to cache Ensembl data into (defaults to clustering code directory)")
    
    args = parser.parse_args()
    
    args.genome_build = args.genome_build.lower()
    
    return args

def main():
    
    args = get_options()
    
    # load all the data
    ensembl = EnsemblRequest(args.cache_dir, args.genome_build)
    mut_dict = load_mutation_rates(args.rates)
    
    old_gene_ids = {}
    # only load the old gene ID converter if we have specified the file
    if args.deprecated_genes_path is not None:
        old_gene_ids = get_deprecated_gene_ids(args.deprecated_genes_path)
    
    known_de_novos = load_de_novos(args.input)
    
    output = open(args.output, "w")
    output.write("gene_id\tmutation_category\tevents_n\tdist\tprobability\n")
    
    iterations = 1000000
    for symbol in sorted(known_de_novos):
        
        de_novos = known_de_novos[symbol]
        
        if len(de_novos["missense"] + de_novos["nonsense"]) < 2:
            continue
        
        # fix HGNC IDs that have been discontinued in favour of other gene IDs
        if symbol in old_gene_ids:
            symbol = old_gene_ids[gene_id]
        
        probs = cluster_de_novos(symbol, de_novos, iterations, ensembl, mut_dict)
        
        if probs is None:
            continue
        
        output.write("{0}\t{1}\t{2}\t{3}\t{4}\n".format(symbol, "missense", \
            len(de_novos["missense"]), probs["miss_dist"], probs["miss_prob"]))
        output.write("{0}\t{1}\t{2}\t{3}\t{4}\n".format(symbol, "nonsense", \
            len(de_novos["nonsense"]), probs["nons_dist"], probs["nons_prob"]))

if __name__ == '__main__':
    main()
